/*
 * jPrologResolver.java
 *
 * Copyright(c) 2002, Particle Corp.
 */

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Stack;

/**
 * jProlog resolution manager
 *
 * @author Alex S.
 * @version 0.0.1
 */
public class jPrologResolver {

    /**
     * database used for resolution
     */
    private Hashtable database;

    /**
     * backtrack stack
     */
    private Stack backtrack;

    /**
     * goals stack
     */
    private Stack goals;

    /**
     * variable hash
     */
    private Hashtable variables;

    /**
     * current variable number
     */
    private long variableNumber;

    /**
     * create a prolog resolver; main class used in prooving goals
     */
    public jPrologResolver(jPrologNode goal,Hashtable db){

        backtrack = new Stack();
        goals = new Stack();
        database = db;
        variableNumber = 0;

        // traverse the list; and place things on stack in reverse order
        // (using a stack to reverse the order of the linked list)
        while(goal != null){
            backtrack.push(goal);
            jPrologNode n = goal.getNext();
            goal.setNext(null);
            goal = n;
        }
        while(!backtrack.isEmpty())
            goals.push(backtrack.pop());

        variables = new Hashtable();
    }

    /**
     * create an instance of a node (instantiate all variables to
     * a temp variable; using a sequential number)
     */
    private jPrologNode createInstanceOfNode(jPrologNode node){
        Hashtable h = new Hashtable();
        if(node.getFunction() == null)
            return createInstanceOfNode(node.cloneNoNext(),h);
        else
            return createDirectInstanceOfNode(node.cloneNoNext());
    }

    /**
     * create a direct; no variable instantiation copy.
     */
    private jPrologNode createDirectInstanceOfNode(jPrologNode node){
        jPrologNode n = null;
        if(node == null)
            return n;
        n = new jPrologNode(node.getName());
        n.setFunction(node.getFunction());
        n.setHead(createDirectInstanceOfNode(node.getHead()));
        n.setNext(createDirectInstanceOfNode(node.getNext()));
        return n;
    }

    /**
     * helper method for instantiation
     */
    private jPrologNode createInstanceOfNode(jPrologNode node,Hashtable h){
        jPrologNode n = null;
        if(node == null)
            return n;
        if(node.isVariable()){
            jPrologNode n1 = (jPrologNode)h.get(node.getName());
            if(n1 == null){
                n = new jPrologNode("_"+(variableNumber++));
                h.put(node.getName(),n);
            }else{
                n = new jPrologNode(n1.getName());
            }
        }else{
            n = new jPrologNode(node.getName());
        }
        n.setHead(createInstanceOfNode(node.getHead(),h));
        n.setNext(createInstanceOfNode(node.getNext(),h));
        return n;
    }

    /**
     * does the actual resolution; returns true if there is a match
     */
    public boolean resolve() throws jPrologException {

        jPrologMatcher matcher = jPrologMatcher.getMatcher();

        // our next goal to prove
        jPrologNode goal = null;

        jPrologList list = null;

//        System.out.println(spaces() + "resolve");

        if(goals.isEmpty()){
            if(backtrack.isEmpty())
                return false;

            while(!backtrack.isEmpty()){

                // if no goals, backtrack
                jPrologBacktrackData btData = (jPrologBacktrackData)backtrack.pop();

                goal = btData.getGoal();
                list = btData.getRule().getNext();   // use next entry

                variables = btData.getVariables();

                for(int i=btData.getNumGoals();i>0;i--)
                    goals.pop();

                goals.push(goal);

//                System.out.println(spaces() + "backtracking: "+goal+"; "+getVarsStr());

                if(list != null)
                    break;
            }

        }else{
            // if there are goals, then get next goal
            goal = (jPrologNode)goals.peek();
            String name = goal.getName() + "/" + goal.getArity();
            list = (jPrologList)database.get(name);
            if(list == null){
                // we've totally failed (undefined term)
                goals.removeAllElements();
                backtrack.removeAllElements();
                throw new jPrologException("undefined predicate: "+name);
            }
        }


//        System.out.println(spaces() + "nextgoal: "+goal);

        while(list != null){

            // if we're here, we know we got a goal, and a list of rules
            while(list != null){

                Hashtable varsBeforeMatch = (Hashtable)variables.clone();

                // create an instance of thing in database
                jPrologNode rule = createInstanceOfNode((jPrologNode)list.getData());

//                System.out.println(spaces()+"applying rule: "+rule);

                Vector replaceRules;

                if(rule.getName().equals(":-")){   // it's a rule
                    rule = rule.getHead();
                    replaceRules = rule.getNext().toList();
                    rule.setNext(null);
                }else{
                    replaceRules = new Vector();
                }

                if(matcher.match(goal,rule,variables) &&
                    (rule.getFunction() == null ||
                       ( rule.getFunction().invoke(variables,database) ))){


//                    System.out.println("rule.getFunction(): "+rule.getFunction());

//                    System.out.println(spaces()+"    matched rule: "+goal+"; "+getVarsStr());

                    goals.pop();    // take off the goal from the list

                    // save backtracking info
                    backtrack.push(new jPrologBacktrackData(list,goal,varsBeforeMatch,replaceRules.size()));

                    for(int i=replaceRules.size()-1;i>=0;i--)
                        goals.push(replaceRules.elementAt(i));

                    // if no more goals, return (every goal matched)
                    if(goals.isEmpty()){
//                        System.out.println(spaces()+"no more goals, found solution.");
                        return true;
                    }

                    // move to next goal

                    // if there are goals, then get next goal
                    goal = (jPrologNode)goals.peek();

//                    System.out.println(spaces() + "nextgoal: "+goal);

                    String name = goal.getName() + "/" + goal.getArity();
                    list = (jPrologList)database.get(name);
                    if(list == null){
                        // we've totally failed (undefined term)
                        goals.removeAllElements();
                        backtrack.removeAllElements();
                        throw new jPrologException("undefined predicate: "+name);
                    }
                }else{
                    variables = varsBeforeMatch;
                    list = list.getNext();
                }
            }

            // didn't find match with this list; backtrack

            while(!backtrack.isEmpty()){

                // if no goals, backtrack
                jPrologBacktrackData btData = (jPrologBacktrackData)backtrack.pop();

                goal = btData.getGoal();
                list = btData.getRule().getNext();   // use next entry

                variables = btData.getVariables();

                for(int i=btData.getNumGoals();i>0;i--)
                    goals.pop();

                goals.push(goal);

//                System.out.println(spaces() + "backtracking: "+goal+"; "+getVarsStr());

                if(list != null)
                    break;
            }
        }
        return false;
    }


    /**
     * construct a variable from other temporary variables.
     */
    public jPrologNode constructVariable(jPrologNode n,Hashtable h){
        if(n == null)
            return null;
        jPrologNode n1,o = n,o1 = n;
        while(n != null && n.isVariable()){
            o = n;
            n = (jPrologNode)h.get(n.getName());
        }
        if(n == null)
            n = o;
        n1 = new jPrologNode(n.getName());
        n1.setHead(constructVariable(n.getHead()));
        n1.setNext(constructVariable(o1.getNext()));
        return n1;
    }


    /**
     * construct a variable from other temporary variables.
     */
    public jPrologNode constructVariable(jPrologNode n){
        return constructVariable(n,variables);
    }

    /**
     * gets list of non-hidden variables
     * after a successful resolution.
     */
    public Hashtable getVariables(){
        Hashtable rv = new Hashtable();
        Enumeration keys = variables.keys();
        while(keys.hasMoreElements()){
            String key = (String)keys.nextElement();
            if(!key.startsWith("_")){
                jPrologNode n = (jPrologNode)variables.get(key);
                rv.put(key,constructVariable(n));
            }
        }
        return rv;
    }

    public String spaces(){
        StringBuffer sb = new StringBuffer();
        int n = backtrack.size();
        for(int i=0;i<n;i++)
            sb.append("    ");
        return sb.toString();
    }

    public String getVarsStr2(){
        StringBuffer rv = new StringBuffer();
        Hashtable v = getVariables();
        Enumeration keys = v.keys();
        while(keys.hasMoreElements()){
            String key = (String)keys.nextElement();
            rv.append(key+"="+v.get(key)+"; ");
        }
        return rv.toString();
    }

    public String getVarsStr(){
        StringBuffer rv = new StringBuffer();
        Enumeration keys = variables.keys();
        while(keys.hasMoreElements()){
            String key = (String)keys.nextElement();
            rv.append(key+"="+variables.get(key)+"; ");
        }
        return rv.toString();
    }


/* bprolog trace

| ?- predecessor(pam,X).
Call: predecessor(pam,_5700a8) ?
  Call: parent(pam,_5700a8) ?
  Exit: parent(pam,bob) ?
Exit: predecessor(pam,bob) ?
X = bob?;
Redo: predecessor(pam,bob) ?
  Redo: parent(pam,bob) ?
  Fail: parent(pam,_5700a8) ?
  Call: parent(pam,_570164) ?
  Exit: parent(pam,bob) ?
  Call: predecessor(bob,_5700a8) ?
    Call: parent(bob,_5700a8) ?
    Exit: parent(bob,ann) ?
  Exit: predecessor(bob,ann) ?
Exit: predecessor(pam,ann) ?
X = ann?;
Redo: predecessor(pam,ann) ?
  Redo: predecessor(bob,ann) ?
    Redo: parent(bob,ann) ?
    Exit: parent(bob,pat) ?
  Exit: predecessor(bob,pat) ?
Exit: predecessor(pam,pat) ?
X = pat?;
Redo: predecessor(pam,pat) ?
  Redo: predecessor(bob,pat) ?
    Redo: parent(bob,pat) ?
    Fail: parent(bob,_5700a8) ?
    Call: parent(bob,_57021c) ?
    Exit: parent(bob,ann) ?
    Call: predecessor(ann,_5700a8) ?
      Call: parent(ann,_5700a8) ?
      Fail: parent(ann,_5700a8) ?
      Call: parent(ann,_5702d4) ?
      Fail: parent(ann,_5702d4) ?
    Fail: predecessor(ann,_5700a8) ?
    Redo: parent(bob,ann) ?
    Exit: parent(bob,pat) ?
    Call: predecessor(pat,_5700a8) ?
      Call: parent(pat,_5700a8) ?
      Exit: parent(pat,jim) ?
    Exit: predecessor(pat,jim) ?
  Exit: predecessor(bob,jim) ?
Exit: predecessor(pam,jim) ?
X = jim?;
Redo: predecessor(pam,jim) ?
  Redo: predecessor(bob,jim) ?
    Redo: predecessor(pat,jim) ?
      Redo: parent(pat,jim) ?
      Fail: parent(pat,_5700a8) ?
      Call: parent(pat,_5702d4) ?
      Exit: parent(pat,jim) ?
      Call: predecessor(jim,_5700a8) ?
        Call: parent(jim,_5700a8) ?
        Fail: parent(jim,_5700a8) ?
        Call: parent(jim,_57038c) ?
        Fail: parent(jim,_57038c) ?
      Fail: predecessor(jim,_5700a8) ?
      Redo: parent(pat,jim) ?
      Fail: parent(pat,_5702d4) ?
    Fail: predecessor(pat,_5700a8) ?
    Redo: parent(bob,pat) ?
    Fail: parent(bob,_57021c) ?
  Fail: predecessor(bob,_5700a8) ?
  Redo: parent(pam,bob) ?
  Fail: parent(pam,_570164) ?
Fail: predecessor(pam,_5700a8) ?
no
*/


}

